@@ -1,63 +1,21 @@
 #include <stdarg.h>
 #include "utils.h"
-/* #include <string.h> */
-#include "version.h"
-#include "H5Zlzo.h"  		       /* Import FILTER_LZO */
-#include "H5Zbzip2.h"  		       /* Import FILTER_BZIP2 */
 
+#define PyString_FromString PyUnicode_FromString
+
+/* See https://numpy.org/doc/1.17/reference/c-api.array.html#c.NO_IMPORT_ARRAY */
+#define NO_IMPORT_ARRAY
+#include <numpy/arrayobject.h>
+
+#ifndef NPY_COMPLEX192
+typedef npy_cdouble npy_complex192;
+#endif
+
+#ifndef NPY_COMPLEX256
+typedef npy_cdouble npy_complex256;
+#endif
 
 /* ---------------------------------------------------------------- */
-
-#ifdef WIN32
-#include <windows.h>
-
-/* This routine is meant to detect whether a dynamic library can be
-   loaded on Windows. This is only way to detect its presence without
-   harming the user.
-*/
-int getLibrary(char *libname) {
-    HINSTANCE hinstLib;
-
-    /* Load the dynamic library */
-    hinstLib = LoadLibrary(TEXT(libname));
-
-    if (hinstLib != NULL) {
-      /* Free the dynamic library */
-      FreeLibrary(hinstLib);
-      return 0;
-    }
-    else {
-      return -1;
-    }
-}
-
-#else  /* Unix platforms */
-#include <dlfcn.h>
-
-/* Routine to detect the existance of shared libraries in UNIX. This
-   has to be checked in MacOSX. However, this is not used right now in
-   utilsExtension.pyx because UNIX does not complain when trying to
-   load an extension library that depends on a shared library that it
-   is not in the system (python raises just the ImportError). */
-int getLibrary(char *libname) {
-    void *hinstLib;
-
-    /* Load the dynamic library */
-    hinstLib = dlopen(libname, RTLD_LAZY);
-
-    if (hinstLib != NULL) {
-      /* Free the dynamic library */
-      dlclose(hinstLib);
-      return 0;
-    }
-    else {
-      return -1;
-    }
-}
-
-
-#endif  /* Win32 */
-
 herr_t set_cache_size(hid_t file_id, size_t cache_size) {
 #if H5_VERS_MAJOR == 1 && H5_VERS_MINOR >= 7
   /* MSVS2005 chokes on declarations after statements */
@@ -84,10 +42,6 @@ herr_t set_cache_size(hid_t file_id, size_t cache_size) {
 
 }
 
-PyObject *_getTablesVersion() {
-  return PyString_FromString(PYTABLES_VERSION);
-}
-
 PyObject *getHDF5VersionInfo(void) {
   long binver;
   unsigned majnum, minnum, relnum;
@@ -103,14 +57,14 @@ PyObject *getHDF5VersionInfo(void) {
   /* A string number */
   if (strcmp(H5_VERS_SUBRELEASE, "")) {
     snprintf(strver, 16, "%d.%d.%d-%s", majnum, minnum, relnum,
-	     H5_VERS_SUBRELEASE);
+             H5_VERS_SUBRELEASE);
   }
   else {
     snprintf(strver, 16, "%d.%d.%d", majnum, minnum, relnum);
   }
 
   t = PyTuple_New(2);
-  PyTuple_SetItem(t, 0, PyInt_FromLong(binver));
+  PyTuple_SetItem(t, 0, PyLong_FromLong(binver));
   PyTuple_SetItem(t, 1, PyString_FromString(strver));
   return t;
 }
@@ -172,71 +126,61 @@ PyObject *createNamesList(char *buffer[], int nelements)
  */
 
 PyObject *get_filter_names( hid_t loc_id,
-			    const char *dset_name)
+                            const char *dset_name)
 {
- hid_t    dset;
- hid_t    dcpl;           /* dataset creation property list */
-/*  hsize_t  chsize[64];     /\* chunk size in elements *\/ */
- int      i, j;
- int      nf;             /* number of filters */
- unsigned filt_flags;     /* filter flags */
- H5Z_filter_t filt_id;       /* filter identification number */
- size_t   cd_nelmts;      /* filter client number of values */
- unsigned cd_values[20];  /* filter client data values */
- char     f_name[256];    /* filter name */
- PyObject *filters;
- PyObject *filter_values;
+  hid_t    dset;
+  hid_t    dcpl;           /* dataset creation property list */
+  /*  hsize_t  chsize[64];     /\* chunk size in elements *\/ */
+  int      i, j;
+  int      nf;             /* number of filters */
+  unsigned filt_flags;     /* filter flags */
+  size_t   cd_nelmts;      /* filter client number of values */
+  unsigned cd_values[20];  /* filter client data values */
+  char     f_name[256];    /* filter name */
+  PyObject *filters;
+  PyObject *filter_values;
 
- /* Open the dataset. */
- if ( (dset = H5Dopen( loc_id, dset_name )) < 0 ) {
-   goto out;
- }
+  /* Open the dataset. */
+  if ( (dset = H5Dopen( loc_id, dset_name, H5P_DEFAULT )) < 0 ) {
+    goto out;
+  }
 
- /* Get the properties container */
- dcpl = H5Dget_create_plist(dset);
- /* Collect information about filters on chunked storage */
- if (H5D_CHUNKED==H5Pget_layout(dcpl)) {
-   filters = PyDict_New();
-    nf = H5Pget_nfilters(dcpl);
-   if ((nf = H5Pget_nfilters(dcpl))>0) {
-     for (i=0; i<nf; i++) {
-       cd_nelmts = 20;
-#if H5_USE_16_API || (H5_VERS_MAJOR == 1 && H5_VERS_MINOR < 7)
-       /* 1.6.x */
-       filt_id = H5Pget_filter(dcpl, i, &filt_flags, &cd_nelmts,
-			       cd_values, sizeof(f_name), f_name);
-#else
-       /* 1.7.x */
-       filt_id = H5Pget_filter(dcpl, i, &filt_flags, &cd_nelmts,
-			       cd_values, sizeof(f_name), f_name, NULL);
-#endif /* if H5_VERSION < "1.7" */
+  /* Get the properties container */
+  dcpl = H5Dget_create_plist(dset);
+  /* Collect information about filters on chunked storage */
+  if (H5D_CHUNKED==H5Pget_layout(dcpl)) {
+    filters = PyDict_New();
+    if ((nf = H5Pget_nfilters(dcpl))>0) {
+      for (i=0; i<nf; i++) {
+        cd_nelmts = 20;
+        H5Pget_filter(dcpl, i, &filt_flags, &cd_nelmts,
+                      cd_values, sizeof(f_name), f_name, NULL);
+        filter_values = PyTuple_New(cd_nelmts);
+        for (j=0;j<(long)cd_nelmts;j++) {
+          PyTuple_SetItem(filter_values, j, PyLong_FromLong(cd_values[j]));
+        }
+        PyMapping_SetItemString (filters, f_name, filter_values);
+      }
+    }
+  }
+  else {
+    /* http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52309 */
+    Py_INCREF(Py_None);
+    filters = Py_None;   /* Not chunked, so return None */
+  }
 
-       filter_values = PyTuple_New(cd_nelmts);
-       for (j=0;j<(long)cd_nelmts;j++) {
-	 PyTuple_SetItem(filter_values, j, PyInt_FromLong(cd_values[j]));
-       }
-       PyMapping_SetItemString (filters, f_name, filter_values);
-     }
-   }
- }
- else {
-   /* http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52309 */
-   Py_INCREF(Py_None);
-   filters = Py_None;  	/* Not chunked, so return None */
- }
+  H5Pclose(dcpl);
+  H5Dclose(dset);
 
- H5Pclose(dcpl);
- H5Dclose(dset);
-
-return filters;
+  return filters;
 
 out:
- H5Dclose(dset);
- Py_INCREF(Py_None);
- return Py_None;  	/* Not chunked, so return None */
-
+  H5Dclose(dset);
+  Py_INCREF(Py_None);
+  return Py_None;        /* Not chunked, so return None */
 }
 
+
 /****************************************************************
 **
 **  get_objinfo(): Get information about the type of a child.
@@ -244,45 +188,123 @@ out:
 ****************************************************************/
 int get_objinfo(hid_t loc_id, const char *name) {
   herr_t     ret;            /* Generic return value         */
-  H5G_stat_t statbuf;
+  H5G_stat_t oinfo;
+  /* H5O_info_t oinfo; H5Oget_info_by_name seems to have performance issues (see gh-402) */
 
   /* Get type of the object, without emiting an error in case the
      node does not exist. */
   H5E_BEGIN_TRY {
-    ret = H5Gget_objinfo(loc_id, name, TRUE, &statbuf);
+    ret = H5Gget_objinfo(loc_id, name, FALSE, &oinfo);
+    /* H5Oget_info_by_name seems to have performance issues (see gh-402) */
+    /*ret = H5Oget_info_by_name(loc_id, name, &oinfo, H5P_DEFAULT);*/
   } H5E_END_TRY;
   if (ret < 0)
     return -2;
-  return statbuf.type;
+  return oinfo.type;
 }
 
 /****************************************************************
 **
-**  gitercb(): Custom group iteration callback routine.
+**  get_linkinfo(): Get information about the type of a link.
 **
 ****************************************************************/
-herr_t gitercb(hid_t loc_id, const char *name, void *data) {
+int get_linkinfo(hid_t loc_id, const char *name) {
+  herr_t     ret;            /* Generic return value         */
+  H5L_info_t linfo;
+
+  /* Get type of the link, without emiting an error in case the
+     node does not exist. */
+  H5E_BEGIN_TRY {
+    ret = H5Lget_info(loc_id, name, &linfo, H5P_DEFAULT);
+  } H5E_END_TRY;
+  if (ret < 0)
+    return -2;
+  return linfo.type;
+}
+
+/****************************************************************
+**
+**  litercb(): Custom link iteration callback routine.
+**
+****************************************************************/
+herr_t litercb(hid_t loc_id, const char *name, const H5L_info_t *info,
+               void *data) {
   PyObject   **out_info=(PyObject **)data;
   PyObject   *strname;
-  herr_t     ret;            /* Generic return value         */
-  H5G_stat_t statbuf;
+  herr_t     ret;
+  H5G_stat_t oinfo;
+  /*H5O_info_t oinfo; H5Oget_info_by_name seems to have performance issues (see gh-402) */
+  int        namedtypes = 0;
 
-    /*
-     * Get type of the object and check it.
-     */
-    ret = H5Gget_objinfo(loc_id, name, FALSE, &statbuf);
-/*     CHECK(ret, FAIL, "H5Gget_objinfo"); */
+  strname = PyString_FromString(name);
 
-    strname = PyString_FromString(name);
-    if (statbuf.type == H5G_GROUP) {
-      PyList_Append(out_info[0], strname);
-    }
-    else if (statbuf.type == H5G_DATASET) {
-      PyList_Append(out_info[1], strname);
-    }
-    Py_DECREF(strname);
+  switch(info->type) {
+    case H5L_TYPE_SOFT:
+    case H5L_TYPE_EXTERNAL:
+      PyList_Append(out_info[2], strname);
+      break;
+    case H5L_TYPE_ERROR:  /* XXX: check */
+      PyList_Append(out_info[3], strname);
+      break;
+    case H5L_TYPE_HARD:
+      /* Get type of the object and check it */
+      ret = H5Gget_objinfo(loc_id, name, FALSE, &oinfo);
+      if (ret < 0)
+        return -1;
 
-    return(0);  /* Loop until no more objects remain in directory */
+      switch(oinfo.type) {
+        case H5G_GROUP:
+          PyList_Append(out_info[0], strname);
+          break;
+        case H5G_DATASET:
+          PyList_Append(out_info[1], strname);
+          break;
+        case H5G_TYPE:
+          ++namedtypes;
+          break;
+        case H5G_UNKNOWN:
+          PyList_Append(out_info[3], strname);
+          break;
+        case H5G_LINK:
+          /* should not happen */
+          PyList_Append(out_info[2], strname);
+          break;
+        default:
+          /* should not happen: assume it is an external link */
+          PyList_Append(out_info[2], strname);
+      }
+
+      /* H5Oget_info_by_name seems to have performance issues (see gh-402)
+      ret = H5Oget_info_by_name(loc_id, name, &oinfo, H5P_DEFAULT);
+      if (ret < 0)
+        return -1;
+
+      switch(oinfo.type) {
+        case H5O_TYPE_GROUP:
+          PyList_Append(out_info[0], strname);
+          break;
+        case H5O_TYPE_DATASET:
+          PyList_Append(out_info[1], strname);
+          break;
+        case H5O_TYPE_NAMED_DATATYPE:
+          ++namedtypes;
+          break;
+        case H5O_TYPE_UNKNOWN:
+          PyList_Append(out_info[3], strname);
+          break;
+        default:
+          / * should not happen * /
+          PyList_Append(out_info[3], strname);
+      }
+      */
+      break;
+    default:
+      /* should not happen */
+      PyList_Append(out_info[3], strname);
+  }
+  Py_DECREF(strname);
+
+  return 0 ;  /* Loop until no more objects remain in directory */
 }
 
 /****************************************************************
@@ -291,20 +313,27 @@ herr_t gitercb(hid_t loc_id, const char *name, void *data) {
 **
 ****************************************************************/
 PyObject *Giterate(hid_t parent_id, hid_t loc_id, const char *name) {
-  int i=0, ret;
-  PyObject  *t, *tdir, *tdset;
-  PyObject *info[2];
+  hsize_t i=0;
+  PyObject  *t, *tgroup, *tleave, *tlink, *tunknown;
+  PyObject *info[4];
 
-  info[0] = tdir = PyList_New(0);
-  info[1] = tdset = PyList_New(0);
+  info[0] = tgroup = PyList_New(0);
+  info[1] = tleave = PyList_New(0);
+  info[2] = tlink = PyList_New(0);
+  info[3] = tunknown = PyList_New(0);
 
-  /* Iterate over all the childs behind loc_id (parent_id+loc_id) */
-  ret = H5Giterate(parent_id, name, &i, gitercb, info);
+  /* Iterate over all the childs behind loc_id (parent_id+loc_id).
+   * NOTE: using H5_INDEX_CRT_ORDER instead of H5_INDEX_NAME causes failures
+   * in the test suite */
+  H5Literate_by_name(parent_id, name, H5_INDEX_NAME, H5_ITER_NATIVE,
+                     &i, litercb, info, H5P_DEFAULT);
 
   /* Create the tuple with the list of Groups and Datasets */
-  t = PyTuple_New(2);
-  PyTuple_SetItem(t, 0, tdir );
-  PyTuple_SetItem(t, 1, tdset);
+  t = PyTuple_New(4);
+  PyTuple_SetItem(t, 0, tgroup);
+  PyTuple_SetItem(t, 1, tleave);
+  PyTuple_SetItem(t, 2, tlink);
+  PyTuple_SetItem(t, 3, tunknown);
 
   return t;
 }
@@ -314,7 +343,8 @@ PyObject *Giterate(hid_t parent_id, hid_t loc_id, const char *name) {
 **  aitercb(): Custom attribute iteration callback routine.
 **
 ****************************************************************/
-static herr_t aitercb( hid_t loc_id, const char *name, void *op_data) {
+static herr_t aitercb( hid_t loc_id, const char *name,
+                       const H5A_info_t *ainfo, void *op_data) {
   PyObject *strname;
 
   strname = PyString_FromString(name);
@@ -331,12 +361,12 @@ static herr_t aitercb( hid_t loc_id, const char *name, void *op_data) {
 **
 ****************************************************************/
 PyObject *Aiterate(hid_t loc_id) {
-  unsigned int i = 0;
-  int ret;
+  hsize_t i = 0;
   PyObject *attrlist;                  /* List where the attrnames are put */
 
   attrlist = PyList_New(0);
-  ret = H5Aiterate(loc_id, &i, (H5A_operator_t)aitercb, (void *)attrlist);
+  H5Aiterate(loc_id, H5_INDEX_CRT_ORDER, H5_ITER_NATIVE, &i,
+             (H5A_operator_t)aitercb, (void *)attrlist);
 
   return attrlist;
 }
@@ -348,15 +378,15 @@ PyObject *Aiterate(hid_t loc_id) {
 **
 ****************************************************************/
 H5T_class_t getHDF5ClassID(hid_t loc_id,
-			   const char *name,
-			   H5D_layout_t *layout,
-			   hid_t *type_id,
-			   hid_t *dataset_id) {
+                           const char *name,
+                           H5D_layout_t *layout,
+                           hid_t *type_id,
+                           hid_t *dataset_id) {
    H5T_class_t  class_id;
    hid_t        plist;
 
    /* Open the dataset. */
-   if ( (*dataset_id = H5Dopen( loc_id, name )) < 0 )
+   if ( (*dataset_id = H5Dopen( loc_id, name, H5P_DEFAULT )) < 0 )
      return -1;
 
    /* Get an identifier for the datatype. */
@@ -380,8 +410,8 @@ H5T_class_t getHDF5ClassID(hid_t loc_id,
 */
 
 PyObject *H5UIget_info( hid_t loc_id,
-			const char *dset_name,
-			char *byteorder)
+                        const char *dset_name,
+                        char *byteorder)
 {
   hid_t       dataset_id;
   int         rank;
@@ -394,9 +424,9 @@ PyObject *H5UIget_info( hid_t loc_id,
   int         i;
 
   /* Open the dataset. */
-  if ( (dataset_id = H5Dopen( loc_id, dset_name )) < 0 ) {
+  if ( (dataset_id = H5Dopen( loc_id, dset_name, H5P_DEFAULT )) < 0 ) {
     Py_INCREF(Py_None);
-    return Py_None;  	/* Not chunked, so return None */
+    return Py_None;     /* Not chunked, so return None */
   }
 
   /* Get an identifier for the datatype. */
@@ -424,7 +454,7 @@ PyObject *H5UIget_info( hid_t loc_id,
   t = PyTuple_New(rank);
   for(i=0;i<rank;i++) {
     /* I don't know if I should increase the reference count for dims[i]! */
-    PyTuple_SetItem(t, i, PyInt_FromLong((long)dims[i]));
+    PyTuple_SetItem(t, i, PyLong_FromLong((long)dims[i]));
   }
 
   /* Release resources */
@@ -463,126 +493,11 @@ out:
  H5Tclose( type_id );
  H5Dclose( dataset_id );
  Py_INCREF(Py_None);
- return Py_None;  	/* Not chunked, so return None */
+ return Py_None;    /* Not chunked, so return None */
 
 }
 
 
-/* Extract a slice index from a PyLong, and store in *pi.  Silently
-   reduce values larger than LONGLONG_MAX to LONGLONG_MAX, and
-   silently boost values less than -LONGLONG_MAX to 0.  Return 0 on
-   error, 1 on success.
-*/
-/* Note: This has been copied and modified from the original in
-   Python/ceval.c so as to allow working with long long values.
-   F. Alted 2005-05-08
-*/
-
-/* Replaced LONLONG_MAX by ll_max because AIX does define the former.
-   F. Alted 2006-10-23
- */
-
-hsize_t _PyEval_SliceIndex_modif(PyObject *v, hsize_t *pi)
-{
-  PY_LONG_LONG ll_max;
-
-  /* I think it should be a more efficient way to know ll_max,
-   but this should work on every platform, be 32 or 64 bits.
-   F. Alted 2005-05-08
-  */
-
-/*  ll_max = (PY_LONG_LONG) (pow(2, 63) - 1); */ /* Works on Unix */
-  ll_max = (PY_LONG_LONG) (pow(2, 62) - 1); /* Safer on Windows */
-
-  if (v != NULL) {
-    PY_LONG_LONG x;
-    if (PyInt_Check(v)) {
-      x = PyLong_AsLongLong(v);
-    }
-    else if (PyLong_Check(v)) {
-      x = PyLong_AsLongLong(v);
-    } else {
-      PyErr_SetString(PyExc_TypeError,
-		      "PyTables slice indices must be integers");
-      return 0;
-    }
-    /* Truncate -- very long indices are truncated anyway */
-    if (x > ll_max)
-      x = ll_max;
-    else if (x < -ll_max)
-      x = -ll_max;
-    *pi = x;
-  }
-  return 1;
-}
-
-/* This has been copied from the Python 2.3 sources in order to get a
-   function similar to the method slice.indices(length) but that works
-   with 64-bit ints and not only with ints.
- */
-
-/* F. Alted 2005-05-08 */
-
-hsize_t getIndicesExt(PyObject *s, hsize_t length,
-		      hsize_t *start, hsize_t *stop, hsize_t *step,
-		      hsize_t *slicelength)
-{
-        /* this is harder to get right than you might think */
-
-        hsize_t defstart, defstop;
-        PySliceObject *r = (PySliceObject *) s;
-
-        if (r->step == Py_None) {
-                *step = 1;
-        }
-        else {
-                if (!_PyEval_SliceIndex_modif(r->step, step)) return -1;
-                if ((PY_LONG_LONG)*step == 0) {
-                        PyErr_SetString(PyExc_ValueError,
-                                        "slice step cannot be zero");
-                        return -1;
-                }
-        }
-
-        defstart = (PY_LONG_LONG)*step < 0 ? length-1 : 0;
-        defstop = (PY_LONG_LONG)*step < 0 ? -1 : length;
-
-        if (r->start == Py_None) {
-                *start = defstart;
-        }
-        else {
-                if (!_PyEval_SliceIndex_modif(r->start, start)) return -1;
-                if ((PY_LONG_LONG)*start < 0L) *start += length;
-                if ((PY_LONG_LONG)*start < 0) *start = ((PY_LONG_LONG)*step < 0) ? -1 : 0;
-                if ((PY_LONG_LONG)*start >= (PY_LONG_LONG)length)
-                        *start = ((PY_LONG_LONG)*step < 0) ? length - 1 : length;
-        }
-
-        if (r->stop == Py_None) {
-                *stop = defstop;
-        }
-        else {
-                if (!_PyEval_SliceIndex_modif(r->stop, stop)) return -1;
-                if ((PY_LONG_LONG)*stop < 0) *stop += length;
-                if ((PY_LONG_LONG)*stop < 0) *stop = -1;
-                if ((PY_LONG_LONG)*stop > (PY_LONG_LONG)length) *stop = length;
-        }
-
-        if (((PY_LONG_LONG)*step < 0 && (PY_LONG_LONG)*stop >= (PY_LONG_LONG)*start)
-            || ((PY_LONG_LONG)*step > 0 && (PY_LONG_LONG)*start >= (PY_LONG_LONG)*stop)) {
-                *slicelength = 0;
-        }
-        else if ((PY_LONG_LONG)*step < 0) {
-                *slicelength = (*stop-*start+1)/(*step)+1;
-        }
-        else {
-                *slicelength = (*stop-*start-1)/(*step)+1;
-        }
-
-        return 0;
-}
-
-
 /* The next provides functions to support a complex datatype.
    HDF5 does not provide an atomic type class for complex numbers
    so we make one from a HDF5 compound type class.
@@ -612,13 +527,13 @@ int is_complex(hid_t type_id) {
       colname1 = H5Tget_member_name(type_id, 0);
       colname2 = H5Tget_member_name(type_id, 1);
       if ((strcmp(colname1, "r") == 0) && (strcmp(colname2, "i") == 0)) {
-	class1 = H5Tget_member_class(type_id, 0);
-	class2 = H5Tget_member_class(type_id, 1);
-	if (class1 == H5T_FLOAT && class2 == H5T_FLOAT)
-	  result = 1;
+        class1 = H5Tget_member_class(type_id, 0);
+        class2 = H5Tget_member_class(type_id, 1);
+        if (class1 == H5T_FLOAT && class2 == H5T_FLOAT)
+          result = 1;
       }
-      free(colname1);
-      free(colname2);
+      H5free_memory(colname1);
+      H5free_memory(colname2);
     }
   }
   /* Is an Array of Complex? */
@@ -662,10 +577,12 @@ static H5T_order_t get_complex_order(hid_t type_id) {
 /* Return the byteorder of a HDF5 data type */
 /* This is actually an extension of H5Tget_order to handle complex types */
 herr_t get_order(hid_t type_id, char *byteorder) {
-  hid_t class_id;
   H5T_order_t h5byteorder;
+  /*
+  hid_t class_id;
 
   class_id = H5Tget_class(type_id);
+  */
 
   if (is_complex(type_id)) {
     h5byteorder = get_complex_order(type_id);
@@ -719,18 +636,86 @@ herr_t set_order(hid_t type_id, const char *byteorder) {
 }
 
 
+/* Create a HDF5 atomic datatype that represents half precision floatting
+   point numbers defined by numpy as float16. */
+hid_t create_ieee_float16(const char *byteorder) {
+  hid_t float_id;
+
+  if (byteorder == NULL)
+    float_id = H5Tcopy(H5T_NATIVE_FLOAT);
+  else if (strcmp(byteorder, "little") == 0)
+    float_id = H5Tcopy(H5T_IEEE_F32LE);
+  else
+    float_id = H5Tcopy(H5T_IEEE_F32BE);
+
+  if (float_id < 0)
+    return float_id;
+
+  if (H5Tset_fields(float_id, 15, 10, 5, 0, 10) < 0)
+    return -1;
+
+  if (H5Tset_size(float_id, 2) < 0)
+    return -1;
+
+  if (H5Tset_ebias(float_id, 15) < 0)
+    return -1;
+
+  return float_id;
+}
+
+
+/* Create a HDF5 atomic datatype that represents quad precision floatting
+   point numbers. */
+hid_t create_ieee_quadprecision_float(const char *byteorder) {
+  hid_t float_id;
+
+  if (byteorder == NULL)
+    float_id = H5Tcopy(H5T_NATIVE_DOUBLE);
+  else if (strcmp(byteorder, "little") == 0)
+    float_id = H5Tcopy(H5T_IEEE_F64LE);
+  else
+    float_id = H5Tcopy(H5T_IEEE_F64BE);
+
+  if (float_id < 0)
+    return float_id;
+
+  if (H5Tset_size(float_id, 16) < 0)
+    return -1;
+
+  if ((H5Tset_precision(float_id, 128)) < 0)
+    return -1;
+
+  if (H5Tset_fields(float_id , 127, 112, 15, 0, 112) < 0)
+    return -1;
+
+  if (H5Tset_ebias(float_id, 16383) < 0)
+    return -1;
+
+  return float_id;
+}
+
+
 /* Create a HDF5 compound datatype that represents complex numbers
    defined by numpy as complex64. */
 hid_t create_ieee_complex64(const char *byteorder) {
   hid_t float_id, complex_id;
 
   complex_id = H5Tcreate(H5T_COMPOUND, sizeof(npy_complex64));
-  if (strcmp(byteorder, "little") == 0)
+  if (byteorder == NULL)
+    float_id = H5Tcopy(H5T_NATIVE_FLOAT);
+  else if (strcmp(byteorder, "little") == 0)
     float_id = H5Tcopy(H5T_IEEE_F32LE);
   else
     float_id = H5Tcopy(H5T_IEEE_F32BE);
-  H5Tinsert(complex_id, "r", HOFFSET(npy_complex64, real), float_id);
-  H5Tinsert(complex_id, "i", HOFFSET(npy_complex64, imag), float_id);
+
+  if (float_id < 0)
+  {
+    H5Tclose(complex_id);
+    return float_id;
+  }
+
+  H5Tinsert(complex_id, "r", 0, float_id);
+  H5Tinsert(complex_id, "i", 4, float_id);
   H5Tclose(float_id);
   return complex_id;
 }
@@ -741,12 +726,85 @@ hid_t create_ieee_complex128(const char *byteorder) {
   hid_t float_id, complex_id;
 
   complex_id = H5Tcreate(H5T_COMPOUND, sizeof(npy_complex128));
-  if (strcmp(byteorder, "little") == 0)
+  if (byteorder == NULL)
+    float_id = H5Tcopy(H5T_NATIVE_DOUBLE);
+  else if (strcmp(byteorder, "little") == 0)
     float_id = H5Tcopy(H5T_IEEE_F64LE);
   else
     float_id = H5Tcopy(H5T_IEEE_F64BE);
-  H5Tinsert(complex_id, "r", HOFFSET(npy_complex128, real), float_id);
-  H5Tinsert(complex_id, "i", HOFFSET(npy_complex128, imag), float_id);
+
+  if (float_id < 0)
+  {
+    H5Tclose(complex_id);
+    return float_id;
+  }
+
+  H5Tinsert(complex_id, "r", 0, float_id);
+  H5Tinsert(complex_id, "i", 8, float_id);
+  H5Tclose(float_id);
+  return complex_id;
+}
+
+
+/* Counterpart for complex192 */
+hid_t create_ieee_complex192(const char *byteorder) {
+  herr_t err = 0;
+  hid_t float_id, complex_id;
+  H5T_order_t h5order = H5Tget_order(H5T_NATIVE_LDOUBLE);
+
+  complex_id = H5Tcreate(H5T_COMPOUND, sizeof(npy_complex192));
+  float_id = H5Tcopy(H5T_NATIVE_LDOUBLE);
+  if (float_id < 0)
+  {
+    H5Tclose(complex_id);
+    return float_id;
+  }
+
+  if ((strcmp(byteorder, "little") == 0) && (h5order != H5T_ORDER_LE))
+    err = H5Tset_order(float_id, H5T_ORDER_LE);
+  else if ((strcmp(byteorder, "big") == 0) && (h5order != H5T_ORDER_BE))
+    err = H5Tset_order(float_id, H5T_ORDER_BE);
+
+  if (err < 0)
+  {
+    H5Tclose(complex_id);
+    return err;
+  }
+
+  H5Tinsert(complex_id, "r", 0, float_id);
+  H5Tinsert(complex_id, "i", 12, float_id);
+  H5Tclose(float_id);
+  return complex_id;
+}
+
+
+/* Counterpart for complex256 */
+hid_t create_ieee_complex256(const char *byteorder) {
+  herr_t err = 0;
+  hid_t float_id, complex_id;
+  H5T_order_t h5order = H5Tget_order(H5T_NATIVE_LDOUBLE);
+
+  complex_id = H5Tcreate(H5T_COMPOUND, sizeof(npy_complex256));
+  float_id = H5Tcopy(H5T_NATIVE_LDOUBLE);
+  if (float_id < 0)
+  {
+    H5Tclose(complex_id);
+    return float_id;
+  }
+
+  if ((strcmp(byteorder, "little") == 0) && (h5order != H5T_ORDER_LE))
+    err = H5Tset_order(float_id, H5T_ORDER_LE);
+  else if ((strcmp(byteorder, "big") == 0) && (h5order != H5T_ORDER_BE))
+    err = H5Tset_order(float_id, H5T_ORDER_BE);
+
+  if (err < 0)
+  {
+    H5Tclose(complex_id);
+    return err;
+  }
+
+  H5Tinsert(complex_id, "r", 0, float_id);
+  H5Tinsert(complex_id, "i", 16, float_id);
   H5Tclose(float_id);
   return complex_id;
 }
@@ -816,7 +874,7 @@ herr_t truncate_dset( hid_t dataset_id,
  if ( (rank = H5Sget_simple_extent_ndims(space_id)) < 0 )
    goto out;
 
- if (rank) {  			/* multidimensional case */
+ if (rank) {    /* multidimensional case */
    /* Book some memory for the selections */
    dims = (hsize_t *)malloc(rank*sizeof(hsize_t));
 
@@ -832,7 +890,7 @@ herr_t truncate_dset( hid_t dataset_id,
    /* Release resources */
    free(dims);
  }
- else {     			/* scalar case (should never enter here) */
+ else {         /* scalar case (should never enter here) */
      printf("A scalar Array cannot be truncated!.\n");
      goto out;
  }
@@ -846,7 +904,31 @@ herr_t truncate_dset( hid_t dataset_id,
 out:
  if (dims) free(dims);
  return -1;
-
 }
 
 
+/*
+ * Helpers for management of HDF5 drivers
+ */
+
+/* DIRECT driver */
+#ifndef H5_HAVE_DIRECT
+
+herr_t pt_H5Pset_fapl_direct(hid_t fapl_id, size_t alignment,
+                             size_t block_size, size_t cbuf_size)
+{
+ return -1;
+}
+
+#endif /* H5_HAVE_DIRECT */
+
+
+/* WINDOWS driver */
+#ifndef H5_HAVE_WINDOWS
+
+herr_t pt_H5Pset_fapl_windows(hid_t fapl_id)
+{
+ return -1;
+}
+
+#endif /* H5_HAVE_WINDOWS */
